/*
TPW AI LOS

Author: TPW 
Line of sight stuff: SaOk 
Azimuth stuff: CarlGustaffa
Additional code: Ollem
Version: 1.04 
Last modified: 20120901
Disclaimer: Feel free to use and modify this code, on the proviso that you post back changes and improvements so that everyone can benefit from them, and acknowledge the original authors in any derivative works.

NOTE: NEEDS BETA 94103 OR GREATER

*/ 

///////////
//VARIABLES
///////////

//STARTUP HINT. 0 = NO HINT, 1 = HINT
tpwlos_hint = 1;

//DEBUGGING. UNITS WITH LOS TO A VISIBLE ENEMY WILL HAVE A BLUE BALL OVER THEIR HEADS. 0 = NO DEBUGGING, 1 = DEBUGGING.
tpwlos_debug = 0;

//MAXIMUM DISTANCE (M). LOS STUFF ONLY WORKS FOR UNITS CLOSER THAN THIS.
tpwlos_maxdist = 100;

//MINIMUM DISTANCE (M). ENEMIES ARE CONSIDERED "VISIBLE" NO MATTER WHAT, IF LESS THAN THIS DISTANCE.
tpwlos_mindist = 25;

//DELAY BEFORE LOS FUNCTIONS START. 
tpwlos_sleep = 1;

	
/////////
//SET UP
/////////

//WAIT
sleep tpwlos_sleep;  

//START HINT
if (tpwlos_hint == 1) then 
	{
	[] spawn 
		{
		hintsilent "TPW AI LOS 1.04 active"; 
		sleep 3; 
		hintsilent ""
		};
	};

	
//////////////////////////////////////////	
//ANGLE OF SUN - ADAPTED FROM CARLGUSTAFFA
//////////////////////////////////////////

tpwlos_fnc_sunangle =
	{
	private ["_lat","_day","_hour"];
	while {true} do 
		{
		_lat = -1 * getnumber(configfile >> "cfgworlds" >> worldname >> "latitude");
		_day = 360 * (datetonumber date);
		_hour = (daytime / 24) * 360;
		tpwlos_sunangle = round (((12 * cos(_day) - 78) * cos(_lat) * cos(_hour)) - (24 * sin(_lat) * cos(_day)));  
		sleep 300; 
		};
	};


////////////////
//MAIN FUNCTION
////////////////

tpwlos_fnc_mainloop =  
	{
	private ["_x","_nexttime","_unit","_near","_nearunits","_a","_b","_dirto","_eyed","_eyepb","_eyepa","_eyedv","_ang","_tint","_lint","_vm","_dist","_esp","_usp","_camo","_formula","_ball"]; 	
	    {
		// ONLY RUN FOR UNINJURED COMBATANTS
		if ((side _x != civilian) && (lifestate _x == "ALIVE")) then 
			{	 
			_unit = _x; 
			_nexttime = _unit getvariable ["tpwlos_nexttime", -1]; 
	
			//SET UP INITIAL UNIT PARAMETERS
			if(_nexttime == -1) then 
				{
				//ENEMY SIDE
				switch (side _unit) do 
				{
				case east: 
					{
					_unit setvariable ["tpwlos_enemyside",west];
					};
				case west: 
					{
					_unit setvariable ["tpwlos_enemyside",east];
					};
				case resistance: 
					{
					_unit setvariable ["tpwlos_enemyside",east];
					};
				};
				
				//DEBUG BLUE BALLS
				if (tpwlos_debug == 1) then 
					{
					_ball = "sign_sphere25cm_ep1" createvehicle getposatl _unit;
					_ball attachto [_unit,[0,0,2.3]]; 
					_unit setvariable ["tpwlos_losball",_ball];
					_ball setObjectTexture [0,"#(argb,8,8,3)color(0.2,0.2,0.9,0.5,ca)"];  
					};
				_unit setvariable ["tpwlos_nexttime", diag_ticktime + random 1];
				};
	
			//IF ENOUGH TIME HAS PASSED SINCE LAST LOS CHECK	
			if(diag_ticktime >= _nexttime) then  
				{	 
				//HIDE DEBUG BALLS
				if (tpwlos_debug == 1) then 
					{
					_ball = _unit getvariable "tpwlos_losball"; 
					_ball hideobject true;
					};		
				_unit setvariable ["tpwlos_nexttime", diag_ticktime + random 1]; 
		
				//FIND NEAR MEN AND CARS
				_nearunits = (getposatl _unit) nearentities [["man","car"],tpwlos_maxdist]; 
					{ 
					_near = _x; 
					tpwlos_cansee = 0; 
					
					//IF ENEMY
					if (side _near == _unit getvariable "tpwlos_enemyside") then  
						//DOES UNIT HAVE LINE OF SIGHT
						{ 
						_a = _unit;  
						_b = _near;  
						_eyedv = eyedirection _a;  
						_eyed = ((_eyedv select 0) atan2 (_eyedv select 1));   
						_dirto = ([_b, _a] call bis_fnc_dirto);  
						_ang = abs (_dirto - _eyed); 
						_eyepa = eyepos _a; 
						_eyepb = eyepos _b; 
						_tint = terrainintersectasl [_eyepa, _eyepb]; 
						_lint = lineintersects [_eyepa, _eyepb]; 
						if (((_ang > 120) && (_ang < 240)) && !(_lint) && !(_tint)) then
							{
							
							//OTHER FACTORS AFFECTING VISIBILITY OF ENEMY
							_vm = (currentvisionmode _unit); 
							if (_vm == 1) then 
								{
								_vm = -1
								} 
								else 
								{
								_vm = 1
								}; 
							_dist = (_unit distance _near); 
							_esp = abs (speed _near);
							_usp = abs (speed _unit);
							_camo = getnumber (configfile >> "cfgvehicles" >> (typeof _near) >> "camouflage");
							
							//GET ENEMY'S CURRENT STANCE - CODE BASED ON CBA_FNC_GETUNITANIM (THANKS OLLEM)
							_anim = toArray(toLower(animationState _near));
							_upos = 1.25; //"other";
							if (count _anim > 11) then
								{
								_dthstr = toString([_anim select 0, _anim select 1, _anim select 2, _anim select 3]);
								_posstr = toString([_anim select 4, _anim select 5, _anim select 6, _anim select 7]);
								_acePosstr = toString([_anim select 8, _anim select 9, _anim select 10, _anim select 11]);
								if (_dthstr == "adth" || _dthstr == "slx_" || _acePosstr == "ppne") then
									{
										_upos = 0.25; //"prone";
									}
								else 
									{
									_upos = switch (_posstr) do
										{
											case "ppne": { 0.25 }; //"prone" 
											case "pknl": { 0.75 }; //"kneel"
											case "perc": { 1.00 }; //"stand"
											default { 0.75 }; //"kneel"
										};
									};
								};
													
							//MAGIC VISIBILITY FORMULA
							_formula = (_vm * (tpwlos_sunangle) * _upos) + (_esp * 6) - (_usp * 2) - _dist + random 40;
							
							//ANYONE LESS THAN 25m IS FAIR GAME
							if (_dist < tpwlos_mindist) then 
								{
								_formula = 200
								};
							
							//IGNORE CAMO ENEMY	
							if (_camo > 0.5) then 
								{
								tpwlos_cansee = _formula
								};
							};
						}; 
					if (tpwlos_cansee > 0) exitwith  
						//IF VISIBLE ENEMY
						{ 
						_unit lookat _near; 
						_unit dofire _near;
						//SHOW DEBUG BALLS
						if (tpwlos_debug == 1) then 
							{
							_ball = _unit getvariable "tpwlos_losball";
							_ball hideobject false;
							}; 
						};
					} foreach _nearunits; 
				};
			}; 
		} foreach allunits;
	};


//////////////////////////////////////////////
//CLEAN UP DEBUG BALLS FROM DEAD/INJURED UNITS
//////////////////////////////////////////////

tpwlos_fnc_cleanup = 
	{
	while {true} do 
		{ 
			//REMOVE DEBUG BALLS FROM DEAD UNITS
			{  
			_unit = _x;  
			_ball = _unit getvariable "tpwlos_losball";
			if ( !(_ball == objNull) ) then		
				{
				detach _ball;   
				deleteVehicle _ball;  
				_unit setvariable ["tpwlos_losball",nil];  
				};	
			} foreach allDead; 
		
			//HIDE DEBUG BALLS FROM INJURED UNITS
			{  
			_unit = _x;  
			_ball = _unit getvariable "tpwlos_losball";
			if (lifestate _unit != "ALIVE") then		
				{
				_ball hideobject true;  
				};	
			} foreach allunits;  		
		sleep 5;
		}; 
	};


////////
//RUN IT
////////

if (tpwlos_debug == 1) then 
	{
	[] spawn tpwlos_fnc_cleanup;
	};

[] spawn tpwlos_fnc_sunangle;	
[tpwlos_fnc_mainloop,0.5] call cba_fnc_addperframehandler;